
def ctmc_datacheck(data, numstates, toltime):
    import numpy as np

    eligiblestates = range(numstates)

    for exid, example in enumerate(data):
        states = example[0]
        times = example[1]

        if not np.all(np.isin(states, eligiblestates)):
            raise Exception(
                ("The example id={:d} has faulty state "
                 "labels/encodings").format(exid))

        if len(np.unique(states)) < 2:
            raise Exception(
                ("The example id={:d} has only 1 distinct "
                 "state").format(exid))

        for i in range(1, len(states)):
            if states[i-1] == states[i]:
                raise Exception(
                    ("The example id={:d} has two consequtive entries "
                     "state[{:d}]==state[{:d}]").format(exid, i-1, i))

        for i, t in enumerate(times):
            if t < toltime:
                raise Exception(
                    ("The example id={:d} has a state[{:d}] that have not "
                     "been active for longer than toltime").format(exid, i))

    return False


def ctmc_errorcheck(transcount, statetime, toltime):
    import numpy as np
    # check transitions counting went wrong
    if np.any(np.diag(transcount) == 0):
        raise Exception(
            ("Transition Count Matrix have diagonal "
             "elements 'm[i,i] != 0'. There are no transition counts "
             "for the i-th state to itself by definition."))

    # check if statetime[i] is big enough to work as divisor
    if np.any(statetime < toltime):
        raise Exception(
            ("The states i="
             + ",".join([str(i) for i in np.where(statetime < toltime)[0]])
             + " have cumulated time period that is smaller than toltime."))

    return False


def ctmc_aggregateevents(data, numstates):
    import scipy.sparse
    import numpy as np

    transcount = scipy.sparse.lil_matrix((numstates, numstates), dtype=int)
    statetime = np.zeros(numstates, dtype=float)

    for _, example in enumerate(data):
        states = example[0]
        times = example[1]

        for i, s in enumerate(states):
            statetime[s] += times[i]
            if i:
                transcount[states[i-1], s] += 1

    return transcount.toarray(), statetime


def ctmc_generatormatrix(transcount, statetime):
    import numpy as np

    tmp = transcount.copy()
    n = tmp.shape[0]

    rowsum = np.sum(tmp, axis=1)
    for i in range(n):
        tmp[i, i] = -rowsum[i]

    genmat = np.zeros(shape=(n, n), dtype=float)
    for i in range(n):
        genmat[i, :] = tmp[i, :] / statetime[i]

    return genmat


def ctmc_fit(data, numstates, transintv=1.0, toltime=1e-8, debug=False):
    """ Continous Time Markov Chain

    Warning:
    --------
    - ctmc_fit assumes a clean data object and does not
        autocorrect any errors as result of it

    The main error sources are

    - transitions counting (e.g. two consequtive states
        has not been aggregated, only one distinct state
        reported) and
    - a state is modeled ore required that does not occur
        in the dataset (e.g. you a certain scale in mind
        and just assume it's in the data) or resp. involved
        in any transition (e.g. an example with just one
        state)

    You can disable any error checking and prevent exceptions
    by setting checks=False but I strongly advice against it.

    Example:
    --------
    Use `ctmc_datacheck` to check during preprocessing the
    dataset

        data = ...
        og.ctmc_datacheck(data, numstates, toltime)

    Disable checks in `ctmc_fit`

        transmat, genmat, transcount, statetime = og.ctmc_fit(
            data, numstates, toltime, checks=False)

    Check aftwards if there has been an error

        og.ctmc_errorcheck(transcount, statetime, toltime)

    """
    import scipy.linalg

    if debug:
        ctmc_datacheck(data, numstates, toltime)  # raises an exception

    transcount, statetime = ctmc_aggregateevents(data, numstates)

    if debug:
        ctmc_errorcheck(transcount, statetime, toltime)  # raises an exception

    genmat = ctmc_generatormatrix(transcount, statetime)

    transmat = scipy.linalg.expm(genmat * transintv)

    return transmat, genmat, transcount, statetime




def ctmc_aggregateevents2(data, numstates, toltime=1e-8):
    import scipy.sparse
    import numpy as np
    import warnings

    transcount = scipy.sparse.lil_matrix((numstates, numstates), dtype=int)
    statetime = np.zeros(numstates, dtype=float)

    for exid, example in enumerate(data):
        states = example[0]
        times = example[1]

        if len(np.unique(states)) < 2:
            warnings.warn(
                ("The example id={:d} has only 1 or no distinct "
                 "state and will be ignored. Please remove incomplete "
                 "examples from the dataset").format(exid))
        else:
            for i, s in enumerate(states):
                if times[i] < toltime:
                    warnings.warn(
                        ("The example id={:d} has a state that "
                         "have not been active for longer than "
                         "toltime.").format(exid))
                else:
                    statetime[s] += times[i]
                    if i:
                        if states[i-1] == s:
                            warnings.warn(
                                ("The example id={:d} contains two entries "
                                 "for the same state in a row. Please  "
                                 "aggregate these obs.").format(exid))
                        else:
                            transcount[states[i-1], s] += 1

    return transcount.toarray(), statetime


def ctmc_generatormatrix2(transcount, statetime, toltime=1e-8):
    """Compute the Generator Matrix

    Parameters:
    -----------
    transcount : ndarray
        NxN matrix with the counted number of transitions
        from state i to j and state j to i (diagonals are
        naturally 0).

    statetime : ndarray
        The cumulated time passed at certain state i.

    Returns:
    --------
    genmat : ndarray
        Generator Matrix

    Correction 1:
    -------------
    The diagonal elements of transcount should be Zero.
    There should be any tr

    Correction 2:
    -------------
    The generator matrix is transition count matrix divided
    by the cumulated time in a certain state

        genmat[i,:] = transcount[i,:] / statetime[i]

    If a certain state is not used, i.e. statetime[i]==0.0,
    then genmat[i,:]=NaN because it's a division by zero.
    Thus, every time

        statetime[i] < toltime

    (e.g. toltime=1e-8) then the generator matrix is reset
    to zero genmat[i,:]=0
    """
    import numpy as np
    import warnings

    # deep copy
    tmp = transcount.copy()

    # get matrix dimension
    n = tmp.shape[0]

    # reset diagonal elements to zero
    # same as "tmp[np.eye(n) == 1] = 0"
    for i in range(n):
        if tmp[i, i] != 0:
            tmp[i, i] = 0
            warnings.warn(
                ("transcount({0:d},{0:d})={1:d} is not Zero. "
                 "There is bug in the way you count transitions"
                 ).format(i, tmp[i, i]))

    # subtract the row sum from the diagonal element
    # same as "tmp -= np.diag(np.sum(tmp, axis=1))"
    rowsum = np.sum(tmp, axis=1)
    for i in range(n):
        tmp[i, i] = -rowsum[i]

    # divide transitions counts by the state idle times
    genmat = np.zeros(shape=(n, n), dtype=float)
    for i in range(n):
        if statetime[i] >= toltime:
            genmat[i, :] = tmp[i, :] / statetime[i]
        else:
            warnings.warn(
                ("The state i={:d} has a cumulated state time "
                 "of less than toltime").format(i))

    # done
    return genmat


def ctmc_fit2(data, numstates, transintv=1.0, toltime=1e-8):
    """ Continous Time Markov Chain (with automatic error correction)

    Warning:
    --------
    - Try ctmc_fit2 if you cannot fix your data object to
        run with ctmc_fit
    - ctmc_aggregateevents2 will ignore all transitions
        with an i-th state that have not been active for
        longer than toltime
    - ctmc_aggregateevents2 will ignore all examples with
        less than 2 distinct states.
    - ctmc_aggregateevents2 will ignore cnt[i,i] transitions
        that does not exist by default.
    - ctmc_generatormatrix2 automagically resets diagonal
        elements of the transcount matrix to zero. There are
        no transitions from the i-th to the i-th state
        by definition.
    - ctmc_generatormatrix2 automagically corrects missing
        cumulated state times (toltime=0) or if measured cumulated
        state time are below a certain threshold (toltime=1e-8 is
        Default), i.e. one state happens to never occur.
        In ctmc_generatormatrix (ctmc_fit) the generator
        matrix would exhibit a NaN-row for the i-th state,
        that never happened.

    In both cases ctmc_fit2 will raise warning messages. If
    you don't like these warning then disable them

        import warnings
        warnings.filterwarnings('ignore')

    """
    import scipy.linalg
    transcount, statetime = ctmc_aggregateevents2(data, numstates, toltime)
    genmat = ctmc_generatormatrix2(transcount, statetime, toltime)
    transmat = scipy.linalg.expm(genmat * transintv)
    return transmat, genmat, transcount, statetime





def yearfrac(a, b):
    """date difference in years between date b and date a."""
    from datetime import date

    # days between date b and a, if a would be in the same year
    if a.day is 29 and a.month is 2:
        diff_days = (b.date() - date(b.year, a.month, 28)).days
    else:
        diff_days = (b.date() - date(b.year, a.month, a.day)).days

    # days of year b
    days_of_yr_b = (date(b.year, 12, 31) - date(b.year-1, 12, 31)).days

    # difference between year b and a, plus the fractional year
    return (b.year - a.year) + diff_days / days_of_yr_b


def table_transform(data, lastdate=None):
    """Transforms array/list to internal data format

    Parameters:
    -----------
    data : ndarray
        Table/panel data with the following columns

            data[:,0]   Examples IDs
            data[:,1]   Event dates as datetime object
            data[:,2]
    """
    import numpy as np

    if lastdate is None:
        lastdate = np.max(data[:, 1])

    newdata = list()

    for _, exampleid in enumerate(np.unique(data[:, 0])):
        # read all entries for the example
        tmp = data[data[:, 0] == exampleid]

        # sort by date
        idxsorted = tmp[:, 1].argsort()
        dates = tmp[idxsorted, 1]
        states = tmp[idxsorted, 2]

        # filter state changes
        idxunique = np.append([True], states[1:] != states[:-1])

        # year fraction
        dates2 = np.append(dates[idxunique], [lastdate])
        times = np.vectorize(yearfrac)(dates2[:-1], dates2[1:])

        # store
        if len(times) > 1:
            newdata.append((list(states[idxunique]), list(times)))

    return newdata



from setuptools import setup


def read(fname):
    import os
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(name='overgang',
      version='0.3.0',
      description='(deprecated package) Markov Modeling',
      long_description=read('README.md'),
      long_description_content_type='text/markdown',
      url='http://github.com/kmedian/overgang',
      author='Ulf Hamster',
      author_email='554c46@gmail.com',
      license='MIT',
      packages=['overgang'],
      install_requires=['numpy', 'datetime', 'scipy'],
      python_requires='>=3',
      zip_safe=False)

